\ TDA is a package for Tri-Dimensional Animation of meshes
\ defined with the functions of the file TDM
 
\ Copyright Paul KOPFF, CLAMART (FRANCE)
 
\ There are two options:
 
\ USE a three color palette (in fact four, the background color
\ is defaulted to black) and flip between two screen pages, using
\ respectively BIT[1..0] and BIT[3..2] planes of VGA memory.
 
\ Thus you can draw parts in RED , GREEN and WHITE (but you can
\ also use other colors instead: just alter the palettes, using
\ the function 2*4-COLORS, from file TGV).
 
\ The second option is true STEREOSCOPIC VIEWING, through colored
\ glasses: You should procure glasses with BLUE and RED filters.
 
\ The following variable controls the differences between the BLUE
\ and the RED IMAGE, thus the tri-dimensional EFFECT
\ It is set for viewing at approximately 40..50 cm off the screen.
 
VARIABLE EFFECT 16 EFFECT !
 
\ This variable controls the speed of animation, through rotation
\ of the point of view around a given axis ( X, Y or Z ).
 
VARIABLE OMEGA
 
\ This variable contains the execution address of the object (or
\ mesh) to display.
 
VARIABLE OBJECT
 
\ This auxiliary vector is useful for stereoscopy: it contains the
\ point-of-view between the eyes. The P vector is in this case
\ flipped from left eye to right eye.
 
3-VECTOR Q
 
\ This is the polar vector defining the rotation speed of the
\ point-of-view. Its length is OMEGA, its direction may be X, Y or Z.
 
3-VECTOR O
 
\ These functions set the O vector to X , Y or Z directions.
\ Each time they are called, they reverse OMEGA; thus calling
\ several times the same function, reverses the rotation.
 
: RX ( omega --- )  0 0         O 3V! ;
 
: RY ( omega --- )  0 0 ROT ROT O 3V! ;
 
: RZ ( omega --- )  0 0 ROT     O 3V! ;
 
\ This function resets the current point-of view and rotation
\ vector to a known state.
 
: RESET 7000 7000 7000 Q 3V! 220 DUP OMEGA ! RX ;
 
\ This function updates the current point-of-view Q
 
: STEP ( --- qx,qy,qz )
         Q 3V@ O 3V@ ROTATION 3VDUP Q 3V! ;
 
 
 
 
 
 
 
\ The following constructs are for selecting the axis of rotation
\ through interactive use of the keys X , Y , and Z.
 
VARIABLE AXIS
 
3 CASE: T-AXIS RX RY RZ ;
 
: R-AXIS OMEGA @ AXIS @ T-AXIS ;
 
: TX OMEGA @ NEGATE OMEGA ! 0 AXIS ! R-AXIS ;
 
: TY OMEGA @ NEGATE OMEGA ! 1 AXIS ! R-AXIS ;
 
: TZ OMEGA @ NEGATE OMEGA ! 2 AXIS ! R-AXIS ;
 
\ The following functions let vary the rotating speed
\ through interactive use of the keys F (faster) and S (slower).
 
5 4 2CONSTANT PO
 
: TF OMEGA @ DUP ABS 500 < IF PO      */ THEN OMEGA ! R-AXIS ;
 
: TS OMEGA @ DUP ABS   5 > IF PO SWAP */ THEN OMEGA ! R-AXIS ;
 
\ The following functions let vary the distance of point-of-view
\ through interactive use of the keys V (farther) and N (nearer).
 
21 20 2CONSTANT PQ
 
: TV R @ 25000  < IF Q 3V@ 3 0 DO PQ      */ ROT LOOP Q 3V! THEN ;
 
: TN R @ S @ 2* > IF Q 3V@ 3 0 DO PQ SWAP */ ROT LOOP Q 3V! THEN ;
 
\ The following function halts the animation: repetitive action
\ on key H enables stepping animation. Any other active key press
\ resumes the continuous movements.
 
: TH BEGIN KEY? UNTIL ;
 
\ This is the active key select function. Any non active key
\ breaks back to FORTH.
 
: KEY-CASE
      KEY? IF
              KEY CASE ASCII x OF TX ENDOF
                       ASCII y OF TY ENDOF
                       ASCII z OF TZ ENDOF
                       ASCII s OF TS ENDOF
                       ASCII f OF TF ENDOF
                       ASCII v OF TV ENDOF
                       ASCII n OF TN ENDOF
                       ASCII h OF TH ENDOF
                                           4 16-COLORS ABORT ENDCASE
            THEN ;
 
\ The following functions initiate the animation.
\ Suppose you have a mesh called FOO : then, only say: AROUND FOO .
 
DEFER AROUND
 
: RGW-AROUND
      ' OBJECT ! RESET
      BEGIN
         PAGE-0 KEY-CASE STEP
         POINT-OF-VIEW OBJECT @ EXECUTE
         PAGE-1 KEY-CASE STEP
         POINT-OF-VIEW OBJECT @ EXECUTE
      AGAIN ;
 
 
 
 
\ The following functions are special for stereoscopy
\ The first one takes the current mean point-of-view (in Q )
\ and computes the increment to add or substract to flip
\ to LEFT-EYE or RIGHT-EYE.
 
: SEPARATION ( qx,qy,qz --- qx,qy,qz,dx,dy,dz,qx,qy,qz,dx,dy,dz )
      3VDUP 3VDUP ROT ROT OR
      IF
         DROP 3VDUP DROP 0 MODULUS >R
         3VDUP MODULUS EFFECT @ / >R DROP NEGATE
         R> R> 2DUP >R >R */ SWAP R> R> */ 0
      ELSE
         DROP EFFECT @ / ROT
      THEN 3VOVER 3VOVER ;
 
: LEFT-VIEW  3V- P 3V! ;
 
: RIGHT-VIEW 3V+ P 3V! ;
 
\ This, finally is the special STEREO-AROUND
 
: STEREO-AROUND
      ' OBJECT ! RESET
      BEGIN
         PAIR-0 KEY-CASE STEP 3VDUP
         POINT-OF-VIEW SEPARATION
         LEFT-EYE  LEFT-VIEW  OBJECT @ EXECUTE
         RIGHT-EYE RIGHT-VIEW OBJECT @ EXECUTE
         PAIR-1 KEY-CASE STEP 3VDUP
         POINT-OF-VIEW SEPARATION
         LEFT-EYE  LEFT-VIEW  OBJECT @ EXECUTE
         RIGHT-EYE RIGHT-VIEW OBJECT @ EXECUTE
      AGAIN ;
 
\ AROUND was DEFERed, and its default behaviour is set to RGW-AROUND.
 
' RGW-AROUND IS AROUND
 
\ That's all Folks !