\ TDM is a package defining a data structure for tridimensional
\ meshes which, when executed draw themselves in a perspective
\ from the current point-of-view.
 
\ Copyright Paul KOPFF, CLAMART (FRANCE)
 
\ works in VGA 16 color mode (MODE 18)
 
18 MODE
 
\ This package uses the TGV package for very fast drawing functions
\ accessing directly the VGA controller (not through BIOS INT 10,
\ which is more than 10 times slower )
 
INCLUDE TGV.FTH
 
\ It uses also the TDV package, where it finds the basic Tri-
\ Dimensional Vector operations it needs.
 
INCLUDE TDV.FTH
 
\ Two variables reminding of the pixel width and heigth of the VGA
\ screens (for scaling of the drawings).
 
VARIABLE HE 320 HE !
VARIABLE VE 240 VE !
 
\ Three vectors defining an orthogonal reference, bound to the
\ current point-of-view, by function POINT-OF-VIEW.
 
3-VECTOR P       \ The point of view
 
3-VECTOR H       \ Horizontal vector, perpendicular to P
 
3-VECTOR V       \ Frontal vector, perpendicular to P and H
 
 
VARIABLE R       \ Length of P , H and V
 
VARIABLE S       \ Radius of a sphere containing the mesh to display
 
 
VARIABLE CRD     \ Number of points defining the mesh
 
VARIABLE CNX     \ Number of segments connecting the points
 
: POINT-OF-VIEW ( xp,yp,zp --- )
        3VDUP P 3V!              \ point-of-view saved in P
        3VDUP MODULUS R !        \ its modulus saved in R
        3VDUP 3VDUP ROT ROT OR   \ is P parallel to Oz ?
        IF                       \ if not, compute H and V
           2DROP 0 MODULUS >R
           3VDUP DROP NEGATE
           R @ R@ */ SWAP
           R @ R@ */ 0 H 3V!
           NEGATE ROT OVER
           R@ */ ROT ROT
           R@ */ R> V 3V!
        ELSE                     \ if so, trivial !
           DROP ROT H 3V!
           NEGATE ROT ROT V 3V!
        THEN ;
 
\ This function sets the P , H and V vectors and the R parameter
\ taking as input a point-of-view vector
 
 
 
 
 
 
: VISION ( x,y,z --- h,v )
        P 3V@ 3VSWAP PERSPECTIVE 3VDUP
        V 3V@ R @ 3V* VE @ DUP ROT S @ */ - >R
        H 3V@ R @ 3V* VE @ S @ */ HE @ + R> ;
 
\ This is the primitive function that transforms a triplet of
\ x,y,z coordinates, into a pair of h,v pixel coordinates by
\ a perspective. It uses also the S parameter and the HE and VE
\ for scaling within the VGA screen.
 
: POINT ( #point --- h,v )
        6 * CRD @ + 3V@ VISION ;
 
\ takes the number of a point of the mesh
 
: SEGMENT ( h,v,adr --- h,v )
        @ DUP ABS POINT ROT 0>
        IF
           2DUP 2ROT LIXEL
        ELSE
           2SWAP 2DROP
        THEN ;
 
\ finds this number at adr in the list of connnexions
\ If positive, draws a line
\ If negative, goes there pen-up
 
: DRAWING ( adr --- )
        DUP CRD !
        DUP 4 + @ S !
        DUP @ 1+ 6 * OVER +
        DUP ROT 2+ @ 2* + SWAP 0 0 2SWAP
        DO I SEGMENT 2 +LOOP
        2DROP ;
 
\ goes through the whole list of connnexions
 
: MESH ( s --- )
        CREATE     4 ALLOT , 0 CRD ! 0 CNX !
        DOES>      DRAWING ;
 
\ is the  defining word that instantiates a mesh; the invocation
\ of a mesh-type-word DOES> the DRAWING .
 
: C ( x,y,z --- )
        , , , 1 CRD +! ;
 
\ is for acquisition of the triplet of coordinates of a point
 
: X ( n --- )
        DUP ABS CRD @ <=
        IF
           , 1 CNX +!
        THEN ;
 
\ is for description of the connexions
 
: END-MESH ( --- )
        LAST @ NAME> >BODY
        CRD @ OVER !
        CNX @ SWAP 2+ ! ;
 
\ ends the instantiation of a mesh.
 
\ After using MESH to give a name to the instance, one should
\ first define all the points using C, then all the connexions
\ using X. This latter function checks if the given points
\ were already defined. See examples in demo file !