\ TGV is a very fast French train
\ Here, the meaning is " TRACES GRANDE VITESSE ",
\ in English: " FAST DRAWINGS "
 
\ DIRECT PROGRAMMING OF THE VGA CONTROLLERS
\ FOR ULTRA-FAST GRAPHIC ANIMATIONS
\ INDIVIDUAL PIXEL SETTING 10 TIMES FASTER
\ THAN WITH THE INT 10 BIOS FUNCTION.
 
\ Copyright Paul KOPFF, CLAMART (FRANCE)
 
BASE @ CONSTANT OLD-BASE
 
\ 1: Management of color coding and palettes
\    We use the BIOS functions because speed is not an issue
 
 
CREATE 16COLBUF 48 ALLOT
 
HEX
 
CODE COLORS> ( ncol,ntab,n --- )
                  CX POP
                  BX POP
              BX  DX MOV
                  BX SHL
              BX  DX ADD
      16COLBUF #  DX ADD
              DS  BX MOV
              BX  ES MOV
                  BX POP
          1017 #  AX MOV
                  10 INT
NEXT
END-CODE
 
\ This function will move n consecutive RGB codes (each on 3 bytes)
\ from position ncol in the 256 Colors table of the VGA controller,
\ to position ntab of the 16COLBUF table.
 
\ Thus, there are two conditions:     ncol + n < 256
\                                     ntab + n < 16
 
\ These conditions are not checked. The words COLORS> and >COLORS
\ should be used only for set-up functions, either with constant
\ behaviour (and then, the programmer should assert the conditions
\ statically) or inside parametrized functions with checks performed
\ before calling.
 
CODE >COLORS ( ncol,ntab,n --- )
                  CX POP
                  BX POP
              BX  DX MOV
                  BX SHL
              BX  DX ADD
      16COLBUF #  DX ADD
              DS  BX MOV
              BX  ES MOV
                  BX POP
          1012 #  AX MOV
                  10 INT
NEXT
END-CODE
 
\ This function is the reverse of the preceeding
 
 
 
 
 
 
CODE 64-COLORS ( nset[0..3] --- )
             0 #  BX MOV
          1013 #  AX MOV
                  10 INT
                  BX POP
              BH  BL XCHG
             3 #  BH AND
             1 #  BL MOV
          1013 #  AX MOV
                  10 INT
NEXT
END-CODE
 
\ This function selects one of four sets of 64 colors.
\ Masking asserts that the number of the set is within [0..3].
 
CODE 16-COLORS ( nset[0..15] --- )
           100 #  BX MOV
          1013 #  AX MOV
                  10 INT
                  BX POP
              BH  BL XCHG
            0F #  BH AND
             1 #  BL MOV
          1013 #  AX MOV
                  10 INT
NEXT
END-CODE
 
\ This function selects one of sixteen sets of 16 colors.
\ Masking asserts that the number of the set is within [0..15].
 
 
 
CODE SAVE-PALETTE ( buf17 --- )
              DS  DX MOV
              DX  ES MOV
                  DX POP
          1007 #  AX MOV
                  10 INT
NEXT
END-CODE
 
\ This function saves a palette into a 17 bytes buffer:
\ 16 colors plus one background default color.
 
CODE IS-PALETTE   ( buf17 --- )
              DS  DX MOV
              DX  ES MOV
                  DX POP
          1002 #  AX MOV
                  10 INT
NEXT
END-CODE
 
\ This function sets a palette from a 17 bytes buffer:
\ 16 colors plus one background default color.
 
 
 
 
 
 
 
 
 
 
 
 
 
 
 
DECIMAL
 
\ The following functions are for setting or retrieving
\ or viewing of RGB color codes
 
: RGB ( n --- paramDO)
     3 * 16COLBUF + DUP 3 + SWAP ;
 
: COLOR! ( nr,nv,nb,n --- )
     RGB DO ROT I C! LOOP ;
 
: COLOR@ ( n --- nr,nv,nb )
     RGB DO I C@ LOOP ;
 
: COLOR? ( n --- display the RGB components )
     RGB DO I C@ . LOOP ;
 
: COLORS? ( display the RGB components of ALL 16 colors )
     16 0 DO CR  I COLOR? LOOP CR ;
 
CREATE NEW_PALETTE  0 C,  1 C,  2 C,  3 C,  4 C,  5 C,  6 C,  7 C,
                    8 C,  9 C, 10 C, 11 C, 12 C, 13 C, 14 C, 15 C,  0 C,
 
CREATE OLD_PALETTE 17 ALLOT
 
: REORGANISE
 
     0 0 6 COLORS>
    20 6 1 COLORS>
     7 7 1 COLORS>
    56 8 8 COLORS>
 
    64 0 16 >COLORS
 
    OLD_PALETTE SAVE-PALETTE ;
 
: DEFAULT
 
    NEW_PALETTE IS-PALETTE 4 16-COLORS ;
 
: STANDARD
 
    OLD_PALETTE IS-PALETTE 0 64-COLORS ;
 
\ The standard predefined  palette uses 16 colors of the first
\ 64 colors map of the VGA controller.
 
\ The remaining 192 RGB cells may be used for alternate palettes
\ They will be divided into 16 colors maps, thus allowing the
\ definition of 12 palettes.
 
\ the first of these 12 palettes is the 16 colors subset of the
\ predefined colors map, which are actually used by the standard
\ palette.
 
REORGANISE DEFAULT
 
\ will do the job
 
\ STANDARD would go back to the initial situation
 
\ WARNING: These three functions work with MY computer, but
\ they might eventually need to be taylored to YOURS !!!
\ See what colors the standard start-up palette uses.
 
 
 
 
 
 
 
 
\ The following 12 bytes tables define the RGB components of
\ two sets of 4 colors.
 
\ The idea is: with the current VGA controllers one cannot,
\ as with some EGA controllers, use TWO pages of screen memory,
\ and thus, one cannot alternate drawing on hidden page then
\ show it and so on... to simulate real-time animation.
 
\ But, if one uses only 4 colors at a time, one can define
\ two palettes, each of which uses 2 of the 4 bit planes
\ avaliable in MODE 18 (VGA 16 colors), and is insensitive
\ to the content of the 2 other planes.
 
\ The setting-up of these palettes is done by 2*4-COLORS,
\ using a 12 bytes table as input for 4 color codes.
 
 
 
\ This table defines pure BLACK RED GREEN and WHITE colors
 
CREATE BRGW     0 C,  0 C,  0 C,  \ BLACK
               42 C,  0 C,  0 C,  \ RED
               21 C, 63 C, 21 C,  \ GREEN
               63 C, 63 C, 63 C,  \ WHITE
 
\ This table defines pure RED and BLUE colors for stereo-
\ scopic viewing through colored glasses.
 
CREATE STEREO   0 C,  0 C,  0 C,  \ BLACK
               63 C,  0 C,  0 C,  \ RED (right eye)
                0 C,  0 C, 63 C,  \ BLUE (left eye)
               63 C,  0 C, 63 C,  \ (  confusion  )
 
: 2*4-COLORS ( nset,table --- )
 
    2DUP
    4 0 DO
        3 0 DO
            DUP I + J 3 * + C@
            J  3 * I + 16COLBUF + DUP 48 + SWAP
               DO DUP I C!
            12 +LOOP DROP
            LOOP
        LOOP DROP
    16 * 0 16 >COLORS
    4 0 DO
        3 0 DO
            DUP I + J 3 * + C@
            J 12 * I + 16COLBUF + DUP 12 + SWAP
               DO DUP I C!
            3  +LOOP DROP
            LOOP
        LOOP DROP
    1+ 16 * 0 16 >COLORS ;
 
\ Symbolic constants for selecting the screen pages
 
5  DUP  CONSTANT 0BRGW
1+      CONSTANT 1BRGW
7  DUP  CONSTANT 0STEREO
1+      CONSTANT 1STEREO
 
\ Set-up of the two pairs of palettes
 
0BRGW   BRGW   2*4-COLORS
0STEREO STEREO 2*4-COLORS
 
 
 
 
 
 
\ 2: The following functions program directly the VGA controller
\    because speed is an issue.
 
HEX
 
CODE WITH ( function --- 0 )
                 3CE #  DX MOV
                   5 #  AX MOV
                    AX  DX OUT   \ WRITING mode 0
 
                   1 #  AX MOV
                    AX  DX OUT   \ with DATA bytes in any plane
 
                        AX POP
                    AX  DX OUT   \ function: SET=0 AND=1 OR=2 XOR=3
 
                FF08 #  AX MOV
                    AX  DX OUT   \ for every bit of a byte
                    AX  AX XOR
                        AX PUSH
NEXT
END-CODE
 
\ This function programs the VGA registers for WRITING mode 0
\ Four options are avaliable:
 
   3 CONSTANT BIT-SET
 803 CONSTANT BIT-AND
1003 CONSTANT BIT-OR
1803 CONSTANT BIT-XOR
 
\ Normally, one should use this function and just after, tell the
\ bit planes on which it should act.
 
\ Thus, the following examples of syntax:
\       BIT-SET WITH BIT-0 BIT-1 PLANES
\       BIT-OR WITH  BIT-3 PLANE
 
\ If one wants to keep the same option but change only the planes,
\ then he should use word ONLY-WITH instead:
\       ONLY-WITH BIT-1 PLANE
 
0 CONSTANT ONLY-WITH
 
\ If one wants to keep the same planes  but change the option,
\ then he should use word SAME like that:
\       BIT-AND WITH SAME
 
: SAME DROP ;
 
: BIT-0  1+ ;
: BIT-1  2+ ;
: BIT-2 4 + ;
: BIT-3 8 + ;
 
CODE PLANE ( mask[0..15] --- )
                 3C4 #  DX MOV
                        AX POP
                  0F #  AX AND
                    AL  AH XCHG
                   2 #  AL MOV
                    AX  DX OUT   \ Selects one or several planes
NEXT
END-CODE           : PLANES PLANE ;
 
\ The PLANES function, for english grammatical compatibility!
 
 
 
 
 
 
CODE CLS ( pattern --- )
                A000 #  AX MOV
                    AX  ES MOV
                    DI  DI XOR
      DECIMAL  19200 #  CX MOV   HEX
                        AX POP
                   REP  AX STOS
NEXT
END-CODE
 
\ This function fills the whole screen with pattern (for blanking,
\ pattern should be 0000 of FFFF).
 
\ The following functions select one page to show and hide the
\ other one. The hidden page is then blanked and prepared for
\ drawing the next image
 
VARIABLE PAGE#
 
DEFER DEFAULT-COLOR
 
: HIDDEN-PAGE       PAGE# @ * PLANE ;
 
: CLEAR-HIDDEN-PAGE BIT-SET WITH BIT-0 BIT-1 HIDDEN-PAGE 0 CLS ;
 
: PAGE-0            0BRGW 16-COLORS  4 PAGE# !
                    CLEAR-HIDDEN-PAGE DEFAULT-COLOR ;
 
: PAGE-1            1BRGW 16-COLORS  1 PAGE# !
                    CLEAR-HIDDEN-PAGE DEFAULT-COLOR ;
 
: RED               BIT-OR WITH BIT-0 HIDDEN-PAGE ;
 
: GREEN             BIT-OR WITH BIT-1 HIDDEN-PAGE ;
 
: WHITE             BIT-OR WITH BIT-0 BIT-1 HIDDEN-PAGE ;
 
' WHITE is DEFAULT-COLOR
 
: PAIR-0            0STEREO 16-COLORS 4 PAGE# ! CLEAR-HIDDEN-PAGE ;
 
: PAIR-1            1STEREO 16-COLORS 1 PAGE# ! CLEAR-HIDDEN-PAGE ;
 
: LEFT-EYE          BIT-OR WITH BIT-0 HIDDEN-PAGE ;
 
: RIGHT-EYE         BIT-OR WITH BIT-1 HIDDEN-PAGE ;
 
\ The following functions set an individual pixel or draw a line
\ defined by pixel coordinates (hence, the name LIXEL).
 
\ After setting the VGA registers, these plotting functions
\ access directly the bit-map memory: this is more than 10 times
\ faster than the INT 10 alternative.
 
CODE PIXEL ( x,y --- )
                        AX  POP
                        CX  POP
                 7FF #  CX  AND
                 7FF #  AX  AND
                 280 #  CX  CMP
0< IF
                 1E0 #  AX  CMP
0< IF
                    AX  BX  MOV
                        BX  SHL
                        BX  SHL
                    AX  BX  ADD
 
 
 
 
 
                  0F #  BH  AND
                  A0 #  BH  OR
                    BX  ES  MOV
                    CX  BX  MOV
                        BX  SHR
                        BX  SHR
                        BX  SHR
                  07 #  CL  AND
                  80 #  AX  MOV
                    AL  CL  ROR
           0  ES:  [BX] AH  MOV
           AL  0  ES:  [BX] MOV
THEN
THEN
NEXT
END-CODE
 
CODE LIXEL ( x1,y1,x2,y2 --- )
                        DX  POP
                        CX  POP
                        BX  POP
                 BX -2 [RP] MOV
                 1   #  AX  MOV
                    BX  DX  SUB
0< IF
                        AX  NEG
                        DX  NEG
THEN
                 AX -4 [RP] MOV
                        BX  POP
                 BX -6 [RP] MOV
                 1   #  AX  MOV
                    BX  CX  SUB
0< IF
                        AX  NEG
                        CX  NEG
THEN
                 AX -8 [RP] MOV
                 0   #  BX  MOV
                    CX  BX  ADD
                    DX  BX  ADD
0= IF
NEXT
THEN
                    CX  DX  CMP
0< IF
                        DX  SHL
                DX -0C [RP] MOV
                    CX  DX  MOV
                    CX  DI  MOV
                        CX  SHL
                CX -0A [RP] MOV
   BEGIN
                -2 [RP] AX  MOV
                -6 [RP] CX  MOV
                -8 [RP] CX  ADD
               -0C [RP] DI  SUB
      0< IF
               -0A [RP] DI  ADD
                -4 [RP] AX  ADD
      THEN
                AX  -2 [RP] MOV
                CX  -6 [RP] MOV
                 7FF #  CX  AND
                 7FF #  AX  AND
                 280 #  CX  CMP
      0< IF
                 1E0 #  AX  CMP
      0< IF
 
 
 
                    AX  BX  MOV
                        BX  SHL
                        BX  SHL
                    AX  BX  ADD
                0F   #  BH  AND
                A0   #  BH  OR
                    BX  ES  MOV
                    CX  BX  MOV
                        BX  SHR
                        BX  SHR
                        BX  SHR
                07   #  CL  AND
                80   #  AX  MOV
                    AL  CL  ROR
           0  ES:  [BX] AH  MOV
           AL  0  ES:  [BX] MOV
      THEN
      THEN
                        DX  DEC
   0= UNTIL
ELSE
                        CX  SHL
                CX -0C [RP] MOV
                    DX  CX  MOV
                    CX  DI  MOV
                        CX  SHL
                CX -0A [RP] MOV
   BEGIN
                -2 [RP] AX  MOV
                -6 [RP] CX  MOV
                -4 [RP] AX  ADD
               -0C [RP] DI  SUB
      0< IF
               -0A [RP] DI  ADD
                -8 [RP] CX  ADD
      THEN
                AX  -2 [RP] MOV
                CX  -6 [RP] MOV
                 7FF #  CX  AND
                 7FF #  AX  AND
                 280 #  CX  CMP
      0< IF
                 1E0 #  AX  CMP
      0< IF
                    AX  BX  MOV
                        BX  SHL
                        BX  SHL
                    AX  BX  ADD
                0F   #  BH  AND
                A0   #  BH  OR
                    BX  ES  MOV
                    CX  BX  MOV
                        BX  SHR
                        BX  SHR
                        BX  SHR
                07   #  CL  AND
                80   #  AX  MOV
                    AL  CL  ROR
           0  ES:  [BX] AH  MOV
           AL  0  ES:  [BX] MOV
      THEN
      THEN
                        DX  DEC
   0= UNTIL
THEN
NEXT
END-CODE        OLD-BASE BASE !          \ That's all, Folks!
 